// Primitives related to librt.vecs that get linked statically
// with compiled modules, instead of being called via a capsule.

#include "vecs_extra_ops.h"

#ifdef MYPYC_EXPERIMENTAL

// All operations are currently implemented as inline functions in vecs_extra_ops.h

#endif // MYPYC_EXPERIMENTAL
